/*************************************************************
 * CSCI P436 - Operating Systems
 *
 * John David Ratliff jdratlif@cs.indiana.edu
 * Chris O'Donnell chodonne@cs.indiana.edu
 *
 * Project 1 - Server Benchmark Client
**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <pthread.h>

#include "sbc.h"

// total bytes received from server
unsigned long int total_bytes = 0;

// protect global data
pthread_mutex_t mutex;

// displays version information
void print_version_info(void) {
	printf("sbc - Server Benchmark Client\n");
	printf("version %s\n", SBC_VERSION);
}

// displays help and syntax screen
void print_help_info(void) {
	print_version_info();
	
	printf("\nsyntax: sbc < args >\n");
	printf("\t-h or --help\n\t\tDisplay this help screen.\n");
	printf("\t-v or --version\n\t\tDisplay version information.\n");
	printf("\t-p or --port\n\t\tThe port to run the server on.\n");
	printf("\t-t or --thread-count\n\t\tThe number of threads to use.\n");
	printf("\t-r or --requests\n\t\tThe number of requests each thread should make.\n");
	printf("\t-s or --server-hostname\n\t\tThe hostname of the server to connect to.\n");
	printf("\n");
}

// parse the command line arguments
void parse_command_line(int argc, char **argv, CLIENT_TRIAL_DATA *client) {
	int loop;
	char *arg;
	
	// setup the client defaults
	client->host_ip = get_ip_address(DEFAULT_HOST);
	client->port = htons((uint16_t)DEFAULT_PORT);
	client->threads = DEFAULT_THREAD_COUNT;
	client->requests = DEFAULT_REQUESTS;
	
	for (loop = 1; loop < argc; loop++) {
		arg = argv[loop];
		
		if ((strcmp(arg,"-h") == 0) || (strcmp(arg,"--help") == 0)) {
			// help
			print_help_info();
			exit(EXIT_SUCCESS);
		} else if ((strcmp(arg,"-v") == 0) || (strcmp(arg,"--version") == 0)) {
			// version info
			print_version_info();
			exit(EXIT_SUCCESS);
		} else if ((strcmp(arg,"-s") == 0) || (strcmp(arg,"--server-hostname") == 0)) {
			// server hostname
			
			if ((++loop) < argc) {
				client->host_ip = get_ip_address(argv[loop]);
			} else {
				fprintf(stderr,"Error: invalid host argument. No host specified!\n");
			}
		} else if ((strcmp(arg,"-p") == 0) || (strcmp(arg,"--port") == 0)) {
			// server port
			
			if ((++loop) < argc) {
				client->port = htons((uint16_t)atoi(argv[loop]));
			} else {
				fprintf(stderr,"Error: invalid port argument. No port specified!\n");
			}
		} else if ((strcmp(arg,"-t") == 0) || (strcmp(arg,"--thread-count") == 0)) {
			// client threads
			
			if ((++loop) < argc) {
				client->threads = atoi(argv[loop]);
			} else {
				fprintf(stderr,"Error: invalid thread count argument. No thread count specified!\n");
			}
		} else if ((strcmp(arg,"-r") == 0) || (strcmp(arg,"--requests") == 0)) {
			// thread requests
			
			if ((++loop) < argc) {
				client->requests = atoi(argv[loop]);
			} else {
				fprintf(stderr,"Error: invalid requests argument. No requests value specified!\n");
			}
		} else {
			// invalid argument
			fprintf(stderr,"Error: invalid argument %s\n",arg);
		}
	}
}

// convert hostname to network byte ordered IP address
in_addr_t get_ip_address(char *hostname) {
	struct hostent *h = gethostbyname(hostname);
	
	if (h == NULL) {
		fprintf(stderr,"Unable to find hostname '%s'\n",hostname);
		return -1;
	}
	
	// return the IP address in network byte format...
	return inet_addr(inet_ntoa(*((struct in_addr *)h->h_addr)));
}

// make a server connection and add the bytes received to the global byte count
short int make_server_connection(in_addr_t ip, uint16_t port) {
	int sockfd, read_count, length;
	long int bytes = 0;
	struct sockaddr_in address;
	char buf[8192];
	int random = rand() % TOTAL_FILES;
	
	// clear the buffer
	memset(&address,0,sizeof(struct sockaddr_in));
	memset(buf,0,8192);
	
	// open the server socket
	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket");
		return FAILURE;
	}
	
	// setup address stuffs
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = ip;
	address.sin_port = port;
	
	// connect to the server
	if (connect(sockfd, (struct sockaddr *)&address, sizeof(struct sockaddr)) == -1) {
		perror("connect");
		return FAILURE;
	}
	
	// format our request with our chosen file
	length = sprintf(buf,"GET /file%d.html HTTP/1.0\r\n\r\n", (random + 1));

#ifdef DEBUG	
	printf("request [%s]\n", buf);
#endif
	
	// send the HTTP request
	if (send(sockfd,buf,length,0) == -1) {
		perror("send");
		close(sockfd);
		return FAILURE;
	}
	
	do {
		// clear the buffer
		memset(buf,0,8192);

		// read response from the socket
		if ((read_count = recv(sockfd,buf,8191,0)) == -1) {
			perror("recv");
			close(sockfd);
			//exit(-1);
			return FAILURE;
		}

		// update the temp byte count
		bytes += read_count;
#ifdef DEBUG
		printf("%s",buf);
#endif
	} while (read_count > 0);

	// close the remote socket
	if (close(sockfd) == -1) {
		perror("close");
	}

	// update the global byte count
	pthread_mutex_lock(&mutex);
	total_bytes += bytes;
	pthread_mutex_unlock(&mutex);

	// return success
	return SUCCESS;
}

void *thread_work(void *param) {
	CLIENT_TRIAL_DATA *client = (CLIENT_TRIAL_DATA *)param;
	short int loop;
	
#ifdef DEBUG
	printf("pthread started thread_work...\n");
#endif

	// make 'client->requests' connections on this thread
	for (loop = 0; loop < client->requests; loop++) {
		if (make_server_connection(client->host_ip, client->port) == FAILURE) {
			printf("failed...\n");
			pthread_exit(NULL);
		}
	}
	
	return param;
}

int main(int argc, char **argv) {
	CLIENT_TRIAL_DATA client;
	pthread_t *threads = NULL;
	short int loop;
	time_t before, after;
	unsigned long int kbs;
	
	// seed the random number generator
	srand(time(NULL));
	
	// parse the command line arguments
	parse_command_line(argc,argv,&client);
	
	// allocate memory for the threads
	if ((threads = (pthread_t *)malloc(client.threads * sizeof(pthread_t))) == NULL) {
		fprintf(stderr,"Fatal error: Insufficient memory to allocate thread space.\n");
		return EXIT_FAILURE;
	}
	
#ifdef DEBUG
	printf("Creating %d threads...\n", client.threads);
#endif

	// get the before time
	before = time(NULL);
	
	// initialize the mutex
	pthread_mutex_init(&mutex,NULL);
	
	// create the physical threads
	for (loop = 0; loop < client.threads; loop++) {
		if (pthread_create((threads + loop),NULL,thread_work,&client)) {
			fprintf(stderr,"Error: Unable to create pthread %d\n", loop);
		}
	}
	
	// wait for threads to finish...
	for (loop = 0; loop < client.threads; loop++) {
		pthread_join(threads[loop], NULL);
	}
	
	// get the time upon completion
	after = time(NULL);
	
#ifdef DEBUG
	printf("threads finished...\n");
#endif

	printf("Total bytes received: %lu\n", total_bytes);
	printf("Total time: ~%lu seconds.\n", (after - before));
	
	// calculate speed
	kbs = (total_bytes / (1024));
	kbs /= ((after - before) > 0) ? (after - before) : 1;

	printf("Average Throughput: %lu KB/s, %lu MB/s\n", kbs, (kbs / 1024));
	
	// destroy the mutex
	pthread_mutex_destroy(&mutex);
	
	// free the thread memory
	free(threads);
	
	return EXIT_SUCCESS;
}

/******************************************************************************
 *  Server Benchmark Client 1.0.0
 *  Copyright (C) 2002 John David Ratliff [jdratlif@cs.indiana.edu]
 *                     Chris O'Donnell    [chodonne@cs.indiana.edu]
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation, 
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************/

