/*************************************************************
 * CSCI P436 - Operating Systems
 *
 * John David Ratliff jdratlif@cs.indiana.edu
 * Chris O'Donnell chodonne@cs.indiana.edu
 *
 * Project 1 - Mini http Server
**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <signal.h>
#include <pthread.h>

#include "mhttpd.h"

// global socket queue and protections
SOCKET_QUEUE *queue;
pthread_mutex_t queue_mutex;
pthread_cond_t queue_full, queue_empty;

// displays version information
void print_version_info(void) {
	printf("mhttpd - the Mini httpd server\n");
	printf("version %s\n", MHTTPD_VERSION);
}

// displays help and syntax screen
void print_help_info(void) {
	print_version_info();
	
	printf("\nsyntax: mhttpd < args >\n");
	printf("\t-h or --help\n\t\tDisplay this help screen.\n");
	printf("\t-v or --version\n\t\tDisplay version information.\n");
	printf("\t-p or --port\n\t\tThe port to run the server on.\n");
	printf("\t-t or --thread-count\n\t\tThe number of threads to use.\n");
	printf("\t-d or --directory\n\t\tThe root http file directory.\n");
	
	printf("\n");
}

// parses the command line arguments into SERVER_DATA format
void parse_command_line(int argc, char **argv, SERVER_DATA *server) {
	int loop;
	char *arg;
	
	// setup the default values
	server->port = DEFAULT_PORT;
	server->workers = DEFAULT_WORKERS;
	
	// the root public_html directory
	memset(server->http_dir,0,1024);
	strcpy(server->http_dir,"/tmp");

	// loop through the argument list
	for (loop = 1; loop < argc; loop++) {
		arg = argv[loop];
		
		if ((strcmp(arg,"-p") == 0) || (strcmp(arg,"--port") == 0)) {
			// server port argument
			
			if ((++loop) < argc) {
				// save the port number
				server->port = atoi(argv[loop]);
			}
		} else if ((strcmp(arg,"-h") == 0) || (strcmp(arg,"--help") == 0)) {
			// help screen
			print_help_info();
			exit(EXIT_SUCCESS);
		} else if ((strcmp(arg,"-v") == 0) || (strcmp(arg,"--version") == 0)) {
			// version info screen
			print_version_info();
			exit(EXIT_SUCCESS);
		} else if ((strcmp(arg,"-t") == 0) || (strcmp(arg,"--thread-count") == 0)) {
			// worker threads to create
			if ((++loop) < argc) {
				// the number of worker threads
				server->workers = atoi(argv[loop]);
				
#ifdef DEBUG
				printf("Worker Thread Count == %d\n",server->workers);
#endif
			}
		} else if ((strcmp(arg,"-d") == 0) || (strcmp(arg,"--directory") == 0)) {
			// root public_html directory
			if ((++loop) < argc) {
				// hope no one has a longer than 1023 character directory... :-)
				strncpy(server->http_dir,argv[loop],1023);
				
#ifdef DEBUG
				printf("public_html dir = [%s]\n", server->http_dir);
#endif
			}
		}
	}
}

// start the server boss on port 'port'
int start_server(int port) {
	int sockfd;
	struct sockaddr_in address;
	int yes = 1;
	
#ifdef DEBUG
	printf("Starting the server on port %d\n", port);
#endif
		
	// create the server socket
	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		fprintf(stderr,"Unable to create the server socket!\n");
		return -1;
	}
	
	// set bind properties
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = INADDR_ANY;
	address.sin_port = htons((unsigned short) port);
	
	// bind the socket
	if (bind(sockfd, (struct sockaddr *)&address, sizeof(struct sockaddr)) == -1) {
		fprintf(stderr,"Unable to bind socket!\n");
		return -1;
	}

	// allow socket rebinding of local ports
	if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(yes)) == -1) {
		fprintf(stderr,"Unable to set reuseable socket binding property...\n");
	}

	// listen for connections
	if (listen(sockfd,10) != 0) {
		fprintf(stderr,"Unable to set socket for listen mode!\n");
		return -1;
	}
	
	// return the socket
	return sockfd;
}

// start running here...
int main(int argc, char **argv) {
	SERVER_DATA server;
	int sockfd;
	pthread_t *workers = NULL;
	short int loop;
	
	// parse command line
	parse_command_line(argc,argv,&server);
	
	// create the queue
	queue = socketqueue_create();
	
	// don't crash on older browers...
	signal(SIGPIPE, SIG_IGN);
	
	// create the mutex and condition variable
	pthread_mutex_init(&queue_mutex,NULL);
	pthread_cond_init(&queue_full,NULL);
	pthread_cond_init(&queue_empty,NULL);

	// attempt to start the server
	if ((sockfd = start_server(server.port)) != -1) {
		// allocate memory for worker thread array
		if ((workers = (pthread_t *)malloc(server.workers * sizeof(pthread_t))) == NULL) {
			// unable to allocate memory for worker threads
			fprintf(stderr,"Fatal error: insufficient memory to create worker threads!\n");
		} else {
#ifdef DEBUG
			printf("building %d worker threads...\n", server.workers);
#endif
			// create worker threads
			for (loop = 0; loop < server.workers; loop++) {
				if (pthread_create((workers + loop), NULL, (void *)worker, server.http_dir)) {
					fprintf(stderr,"unable to create worker thread %d\n",loop);
				}
			}
		}
		
		// do the boss work
		boss(&sockfd);
				
#ifdef DEBUG
		printf("Server shutdown in progress...\n");
#endif

		for (loop = 0; loop < server.workers; loop++) {
#ifdef DEBUG
			printf("Cancelling worker thread %d\n", loop);
#endif
			pthread_cancel(workers[loop]);
		}
		
#ifdef DEBUG
		printf("Closing the server socket...\n");
#endif

		// close the server socket
		close(sockfd);
	}

	// exit the main thread
	pthread_exit(NULL);
	
	return EXIT_SUCCESS;
}

/******************************************************************************
 *  Mini http server 1.0.0
 *  Copyright (C) 2002 John David Ratliff [jdratlif@cs.indiana.edu]
 *                     Chris O'Donnell    [chodonne@cs.indiana.edu]
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation, 
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************/

