/*************************************************************
 * CSCI P436 - Operating Systems
 *
 * John David Ratliff jdratlif@cs.indiana.edu
 * Chris O'Donnell chodonne@cs.indiana.edu
 *
 * Project 1 - Mini http Server
**************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/socket.h>
#include <sys/stat.h>

#include "mhttpd.h"

// our error pages
static const char error_400[] = "<html><head><title>Error 400: Bad Request</title></head><body><h2>Illegal Request</h2>Illegal URI.</body></html>";
static const char error_404[] = "<html><head><title>Error 404: File Not Found</title></head><body><h2>File not found</h2>Please try your link again.</body></html>";
static const char error_403[] = "<html><head><title>Error 403: Forbidden</title></head><body><h2>Forbidden URI</h2>Sorry, you don't have enough clearance for this URI.</body></html>";
static const char error_500[] = "<html><head><title>Error 500: Internal Server Error</title></head><h2>Server Error</h2>This could be almost anything. Sorry to be so vague.</body></html>";
static const char error_501[] = "<html><head><title>Error 501: Method Not Supported Error</title></head><h2>Not a GET Request</h2>This server supports only HTTP/1.0 GET</body></html>";

void send_http_header(int sockfd, int status, const char *reason, unsigned long int length, const char *content_type, time_t last_modified) {
	char buffer[1024], time_buffer[50];
	int string_length;
	time_t now = time(NULL);
	
	// send the status response code
	string_length = sprintf(buffer,"HTTP/1.0 %d %s\r\n",status,reason);
	send(sockfd, buffer, string_length, 0);
	
#ifdef DEBUG
	printf("%s",buffer);
#endif
	
	// send the HTTP header data
	
	// send the Date header
	strftime(time_buffer, 50, HTTP_DATE_FMT, gmtime(&now));
	string_length = sprintf(buffer, "Date: %s\r\n", time_buffer);
	send(sockfd, buffer, string_length, 0);
	
#ifdef DEBUG
	printf("%s",buffer);
#endif
	
	// send the Server header
	send(sockfd, "Server: mhttpd 1.0.0\r\n", 22, 0);
	
#ifdef DEBUG
	printf("Server: mhttpd 1.0.0\r\n");
#endif
	
	// send the Content-Length header
	string_length = sprintf(buffer, "Content-Length: %lu\r\n", length);
	send(sockfd, buffer, string_length, 0);
	
#ifdef DEBUG
	printf("%s",buffer);
#endif
	
	// send the Content-Type header
	string_length = sprintf(buffer, "Content-Type: %s\r\n", content_type);
	send(sockfd, buffer, string_length, 0);
	
#ifdef DEBUG
	printf("%s",buffer);
#endif
	
	// send the Last-Modified header
	strftime(time_buffer, 50, HTTP_DATE_FMT, gmtime(&last_modified));
	string_length = sprintf(buffer, "Last-Modified: %s\r\n", time_buffer);
	send(sockfd, buffer, string_length, 0);
	
#ifdef DEBUG
	printf("%s",buffer);
#endif
	
	// send the extra CRLF as notice that the header is complete
	send(sockfd, "\r\n", 2, 0);
}

void send_error_page(int sockfd, int error_code, const char *reason, const char *page) {
	int length = strlen(page);
	
	send_http_header(sockfd, error_code, reason, length, "text/html", time(NULL));
	send(sockfd, page, length, 0);
	close(sockfd);
}

int get_next_line(const char *strings, char *buffer, int max_length) {
	int loop;
	int string_length = strlen(strings);
	
	if (string_length > max_length) {
		string_length = max_length;
	}
	
	for (loop = 0; loop < string_length; loop++) {
		if ((strings[loop] == '\n') || (strings[loop] == '\r')) {
			break;
		}
	}
	
	strncpy(buffer,strings,loop);
	
	return loop;
}

void parse(int sockfd, const char *http_dir, const char *request) {
	char method[1024], uri[1024], protocol[1024];
	char buffer[10000];
	const char *filename;
	struct stat file_stats;
	int length;
	FILE *f;
	
	// erase the buffer
	memset(buffer,0,10000);

	// get the first line of the request
	get_next_line(request,buffer,10000);
	
	if (sscanf(buffer, "%[^ ] %[^ ] %[^ ]", method, uri, protocol) != 3) {
		send_error_page(sockfd, 400, "Bad Request", error_400);
		return;
	}
	
	if (strcmp(method, "GET") != 0) {
		send_error_page(sockfd, 501, "Not Implemented", error_501);
		return;
	}
	
	// remove the leading slash from the file path
	filename = (uri + 1);
	
	// if we have an empty path
	if (filename[0] == 0) {
		filename = "index.html";
	}
	
	// check for invalid characters in the file path
	if ((filename[0] == '/') || (strcmp(filename,"..") == 0) ||
	    (strstr(filename,"../") != NULL) || (strstr(filename,"/..") != NULL)) {
		send_error_page(sockfd, 400, "Bad Request", error_400);
		return;
	}
	
	if (chdir(http_dir) == -1) {
		send_error_page(sockfd, 404, "File Not Found", error_404);
		return;
	}
	
	if (stat(filename, &file_stats) == -1) {
		send_error_page(sockfd, 404, "File Not Found", error_404);
		return;
	}
	
	// don't handle directories, yet
	if (S_ISDIR(file_stats.st_mode)) {
		send_error_page(sockfd, 500, "Internal Server Error", error_500);
		return;
	}
	
	if ((f = fopen(filename, "rt")) == NULL) {
		send_error_page(sockfd, 403, "Forbidden", error_403);
		return;
	}
	
	// we have a good file, send the headers
	send_http_header(sockfd, 200, "Ok", file_stats.st_size, "text/html", file_stats.st_mtime);
	
	while (!feof(f)) {
		length = fread(buffer, sizeof(char), 10000, f);
		buffer[length] = 0;
		send(sockfd, buffer, length, 0);
	}
	
	fclose(f);
	
	close(sockfd);
}
    
/******************************************************************************
 *  Mini http server 1.0.0
 *  Copyright (C) 2002 John David Ratliff [jdratlif@cs.indiana.edu]
 *                     Chris O'Donnell    [chodonne@cs.indiana.edu]
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation, 
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************/

